"""Asynchronous Python client for Sonoff DIY devices."""
import asyncio
import json
import socket
from typing import Dict, Optional, Union

import aiohttp
import async_timeout
from yarl import URL

from .__version__ import __version__
from .const import ERROR_CODES, STATE_OFF, STATE_ON, STATE_RESTORE, STATE_STAY
from .exceptions import SonoffDIYConnectionError, SonoffDIYError
from .models import Info


class SonoffDIY:
    """Main class for handling connections with a Sonoff DIY device."""

    info: Optional[Info] = None

    def __init__(
        self,
        host: str,
        device_id: str = "",
        loop: asyncio.events.AbstractEventLoop = None,
        port: int = 8081,
        request_timeout: int = 3,
        session: aiohttp.client.ClientSession = None,
        user_agent: str = None,
    ) -> None:
        """Initialize connection with a Sonoff DIY device."""
        self._loop = loop
        self._session = session
        self._close_session = False

        self.device_id = device_id
        self.host = host
        self.port = port
        self.request_timeout = request_timeout
        self.user_agent = user_agent

        if user_agent is None:
            self.user_agent = f"PythonSonoffDIY/{__version__}"

    async def _request(
        self, uri: str = "", data: Optional[Dict[str, Union[int, str]]] = None,
    ) -> Optional[Dict[str, Union[bool, int, str]]]:
        """Handle a request to a Sonoff DIY device."""
        url = URL.build(
            scheme="http", host=self.host, port=self.port, path="/zeroconf/"
        ).join(URL(uri))

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._session is None:
            self._session = aiohttp.ClientSession(loop=self._loop)
            self._close_session = True

        try:
            with async_timeout.timeout(self.request_timeout):
                response = await self._session.request(
                    "POST",
                    url,
                    json={"deviceid": self.device_id, "data": data or {}},
                    headers={
                        "User-Agent": self.user_agent,
                        "Accept": "application/json, */*",
                    },
                )
                response.raise_for_status()
        except asyncio.TimeoutError as exception:
            raise SonoffDIYConnectionError(
                "Timeout occurred while connecting to Sonoff DIY device"
            ) from exception
        except (
            aiohttp.ClientError,
            aiohttp.ClientResponseError,
            socket.gaierror,
        ) as exception:
            raise SonoffDIYConnectionError(
                "Error occurred while communicating with Sonoff DIY device"
            ) from exception

        content_type = response.headers.get("Content-Type", "")
        if "application/json" not in content_type:
            text = await response.text()
            raise SonoffDIYError(
                "Unexpected response from Sonoff DIY device",
                {"Content-Type": content_type, "response": text},
            )

        response_data = await response.json()
        if "error" not in response_data:
            raise SonoffDIYError(
                "Unexpected response from Sonoff DIY device", response_data
            )

        if response_data["error"] != 0:
            error = ERROR_CODES.get(response_data["error"], "Unknown error occurred")
            raise SonoffDIYError(error, response_data)

        if "data" not in response_data:
            return None

        # itead doesn't know how JSON works apparently, so they might return
        # it packed inside the response as a serialized string.
        if isinstance(response_data["data"], str):
            return json.loads(response_data["data"])

        return response_data["data"]

    async def update_info(self) -> Optional[Info]:
        """Get all information about the Sonoff DIY device."""
        try:
            data = await self._request("info")
            signal_strength = await self._request("signal_strength")
        except SonoffDIYError as exception:
            self.info = None
            raise exception

        if data is None:
            self.info = None
            raise SonoffDIYError("Did not receive data from Sonoff DIY device")

        if signal_strength is not None:
            data.update(signal_strength)

        self.info = Info.from_dict(self.device_id, data)
        return self.info

    async def turn_on(self) -> None:
        """Turn on Sonoff DIY device."""
        await self._request("switch", {"switch": STATE_ON})

    async def turn_off(self) -> None:
        """Turn off Sonoff DIY device."""
        await self._request("switch", {"switch": STATE_OFF})

    async def pulse_on(self, pulse_width: Optional[int] = None) -> None:
        """Turn on inching on Sonoff DIY device."""
        data: Dict[str, Union[int, str]] = {"pulse": STATE_ON}
        if pulse_width is not None:
            data["pulse_width"] = pulse_width

        await self._request("pulse", data)

    async def pulse_off(self) -> None:
        """Turn off inching on Sonoff DIY device."""
        await self._request("pulse", {"pulse": STATE_OFF})

    async def pulse_width(self, pulse_width: int) -> None:
        """Set pulse width for Sonoff DIY device."""
        await self.update_info()
        pulse = STATE_ON if self.info and self.info.pulse else STATE_OFF
        await self._request("pulse", {"pulse": pulse, "pulse_width": pulse_width})

    async def wifi(self, ssid: str, password: str) -> None:
        """Set WiFi network for the Sonoff DIY device to connect to."""
        await self._request("wifi", {"ssid": ssid, "password": password})

    async def ota_unlock(self) -> None:
        """Unlock Sonoff DIY device for OTA updates."""
        await self._request("ota_unlock")

    async def ota_flash(self, url: str, sha: str) -> None:
        """OTA flash firmware to the Sonoff DIY device."""
        await self._request("ota_flash", {"downloadUrl": url, "sha256sum": sha})

    async def power_on_state(self, state: str) -> None:
        """Set state of switch when powering up."""
        if state not in (STATE_ON, STATE_OFF, STATE_RESTORE, STATE_STAY):
            raise SonoffDIYError("Invalid startup value, accepted: on, off, restore")
        # Translate restore back to stay, which is weird termonology
        if state == STATE_RESTORE:
            state = STATE_STAY
        await self._request("startup", {"startup": state})

    async def close(self) -> None:
        """Close open client session."""
        if self._session and self._close_session:
            await self._session.close()

    async def __aenter__(self) -> "SonoffDIY":
        """Async enter."""
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        await self.close()



"""Models for Sonoff DIY devices."""

ERROR_CODES = {
    400: "Request body is not valid JSON",
    401: "Unauthorized request",
    403: "OTA function is locked",
    404: "Device does not support requested Device ID",
    408: "Timeout while downloading firmware",
    413: "OTA firmware exceeds device allowed size limit",
    422: "Invalid request parameters",
    424: "Firmware download failed",
    471: "Firmware integrity check failed",
    500: "Device ID or API key is not authorized by vendor's OTA service",
    503: "Could not connect to vendor's OTA unlock service",
}

STATE_OFF = "off"
STATE_ON = "on"
STATE_RESTORE = "restore"
STATE_STAY = "stay"



"""Models for Sonoff DIY devices."""

import attr

from .const import STATE_OFF, STATE_ON, STATE_RESTORE, STATE_STAY


@attr.s(auto_attribs=True, frozen=True)
class Info:
    """Object holding information and states of the Sonoff DIY device."""

    device_id: str
    on: bool
    ota_unlock: bool
    power_on_state: str
    pulse_width: int
    pulse: bool
    signal_strength_percentage: int
    signal_strength: int
    ssid: str

    @staticmethod
    def from_dict(device_id: str, data: dict):
        """Return device object from Sonoff DIY API response."""

        signal_strength = data.get("signalStrength", -100)
        if signal_strength <= -100:
            signal_strength_percentage = 0
        elif signal_strength >= -50:
            signal_strength_percentage = 100
        else:
            signal_strength_percentage = 2 * (signal_strength + 100)

        power_on_state = data.get("startup", STATE_OFF)
        if power_on_state == STATE_STAY:
            power_on_state = STATE_RESTORE

        return Info(
            device_id=device_id,
            on=data.get("switch") == STATE_ON,
            ota_unlock=data.get("otaUnlock", False),
            power_on_state=power_on_state,
            pulse_width=data.get("pulseWidth", 0),
            pulse=data.get("pulse") == STATE_ON,
            signal_strength_percentage=signal_strength_percentage,
            signal_strength=signal_strength,
            ssid=data.get("ssid", ""),
        )



#!/usr/bin/env python
"""The setup script."""
import os
import re
import sys

from setuptools import find_packages, setup


def get_version():
    """Get current version from code."""
    regex = r"__version__\s=\s\"(?P<version>[\d\.]+?)\""
    path = ("sonoffdiy", "__version__.py")
    return re.search(regex, read(*path)).group("version")


def read(*parts):
    """Read file."""
    filename = os.path.join(os.path.abspath(os.path.dirname(__file__)), *parts)
    sys.stdout.write(filename)
    with open(filename, encoding="utf-8", mode="rt") as fp:
        return fp.read()


with open("README.md") as readme_file:
    readme = readme_file.read()

setup(
    author="Franck Nijhof",
    author_email="opensource@frenck.dev",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    description="Asynchronous Python client for Sonoff DIY devices.",
    include_package_data=True,
    install_requires=["aiohttp>=3.0.0", "attrs>=19.0.0", "yarl"],
    keywords=["sonoff", "diy", "api", "async", "client"],
    license="MIT license",
    long_description_content_type="text/markdown",
    long_description=readme,
    name="sonoffdiy",
    packages=find_packages(include=["sonoffdiy"]),
    test_suite="tests",
    url="https://github.com/frenck/python-sonoffdiy",
    version=get_version(),
    zip_safe=False,
)
